# Information #################################################

# Time zone and hour format: Hour of the day is shown in UTC. Value at Hour 1 correspond to the hour starting at 1 and ending at 2. As April falls in summer time (UTC + 2 hours), to convert from local time to UTC we move two hours backwards. So, 7 am local time is 5 am UTC.   
# Observations source: gene/ghost
# Models intercompared:  
# Raw model from Caliope
# LOOCV applied to Caliope


# Measure the time it takes to run the analysis
start_time <- Sys.time()

# Set libraries ##############################################
# ! check the version of the package or R if any library or function does not work correctly

library(urbanR)
library(RNetCDF)
library(sp)
library(sf)
library(ncdf4)
library(rgeos) 
library(ggplot2)
library(plyr) #OJO! la llibreria plyr ha de ser cridada abans que dplyr 
library(dplyr) 
library(broom)
library(magrittr)
library(ggpubr)
library(raster)
library(rgdal)
library(gstat)
library(sp)
library(abind)
library(openair)
library(tidyr)
library(gbm)
library(dplyr)
library(lubridate)



# Load settings ###############################################
## LOAD config_file
args = commandArgs(TRUE)
if (length(args) > 0) {
  configFile_path <- args[1]
} else {
  cat("\n***WARNING: no ConfigFile was provided. Using default ConfigFile.\n")
}
#config_file it's a file where we change the data, where we want to save the results, etc. so now we extract these data from it:
# configFile_path <- "/esarchive/scratch/ccarnere/RScripts/results_regional_2023/results_gene_o3/config_file_variogram.yml"
# configFile_path <- "/esarchive/scratch/ccarnere/RScripts/results_regional_2018/results_gene_pm2p5/config_file.yml"
# configFile_path <- "/esarchive/scratch/ccarnere/RScripts/results_regional_2018/results_gene_pm10/config_file.yml"

config          <- yaml::yaml.load_file(configFile_path)

# Read settings from config_file ##################################
stations_path   <- config$stations_path
obs_path        <- config$obs_path
model_path      <- config$model_path

read_data_gbm   <- config$read_data_gbm #path per cridar els resultats d'anys anteriors del UK per entrenar el gbm
variogram_path  <- config$variogram_path
path_code       <- config$path_code

# Pollutant       	 
pollutant       <- config$pollutant 
print(paste0("Selected pollutant: ", pollutant))

# Spatial  Domain   
lon_domain           <- config$lon_domain # lon max, lon min
lat_domain           <- config$lat_domain # lat max, lat min

# Coordinates   	 
compute_in_UTM         <- config$compute_in_UTM
crs_latlon             <- config$crs_latlon
crs_utm                <- config$crs_utm

# Temporal Domain
time_resolution       <- config$time_resolution # hourly or daily
model_time_resolution <- config$model_time_resolution # time resolution of the model : hoyrly or daily
month_names           <- config$month_names
file_format           <- config$file_format # ghost or gene

years_gbm             <- config$years_gbm #years which we want the data to be readed by gbm to get trained

valid_formats <- c("gene", "ghost")

if(time_resolution == "hourly" && file_format %in% valid_formats){
  print(paste0("The model and observation are in hourly format, using ", file_format, " file format"))
} else if(time_resolution == "daily" && file_format %in% valid_formats){
  print(paste0("The model and observations are in daily format, using ", file_format, " file format"))
} else{
  stop("Wrong time_resolution or file_format. Please check the configuration.")
}

date_time_begin <- config$date_time_begin
date_time_end   <- config$date_time_end

# EXECUTIONS

set_stations              <- config$set_stations
set_observations          <- config$set_observations
set_stations_availability <- config$set_stations_availability
set_model                 <- config$set_model

execute_ok                <- config$execute_ok
execute_gbm               <- config$execute_gbm #here we obtain the observations with gap filling

execute_uk_gbm            <- config$execute_uk_gbm 
execute_uk_loocv          <- config$execute_uk_loocv
execute_uk                <- config$execute_uk
execute_loocv             <- config$execute_loocv

execute_statistics        <- config$execute_statistics
execute_general_st        <- config$execute_general_st
execute_temporal_var      <- config$execute_temporal_var
execute_time_series       <- config$execute_time_series


# Save Data 
read_data_gbm              <- config$read_data_gbm

save_result                <- config$save_result
save_model                 <- config$save_model
save_obs                   <- config$save_obs
save_OK                    <- config$save_OK
save_gbm                   <- config$save_gbm
save_LOOCV                 <- config$save_LOOCV
save_stations              <- config$save_stations

save_general_statistics    <- config$save_general_statistics 
save_time_series           <- config$save_time_series 
save_temporal_variation    <- config$save_temporal_variation 


# Auxiliar Variograms
# As we had problems with Ordinary and Universal Kriging variograms, we saved one of each to read them in case we have problems with the one executed for a specific date or hour
#NOTE: in principle, if it was from the gene or ghost, it would not affect the variogram, rather than the contaminant itself. But in this case, we will save one for each.
variogram_ok <- paste0(variogram_path, "Variogram_OK_",file_format,"_",pollutant,".rds") 
variogram_uk <- paste0(variogram_path, "Variogram_UK_",file_format,"_",pollutant,".rds")


# import functions #############################################
source(paste0(path_code,"utils.R"))



# Set Units ####################################################
print("Setting units of obs and model")
# We need to adjust the units of the observations to ug/m3, so first we pick up the constant depending on the pollutant that we are working with
pol          <- c("no2","o3","pm10","pm2p5","so2","co","no")
k            <- c(1.881, 1.963, 1, 1, 2.6609, 1.15, 1.227) # ppb to ug/m3, except for CO: ppm to mg/m3

# We will use the limit parameter to avoid observations that we consider wrong in loocv loop. As for NO2 the alert is established at 400 ug/m3, and 240 for O3 we will put this number as the boundary
limit        <- c(1000, 400, 500, 200, 1000, 5000, 1000) #the limit for pm10 is 50 ug/m3, 25 ug/m3, but i put 100 just in case 

position     <- match(pollutant, pol) 
number       <- match(pol[position], pol)

limit        <- limit[number]
obs_constant <- k[number]

#to arrange set_stations or observations, we need to configure name_pollutant, based in how the script name with the data is saved
if(pollutant %in% c("no2","o3","so2","co")){
  name_pollutant   <- paste0("sconc",pollutant)
  mod_constant     <- k[number]*1e03
  min_obs          <- 20 #min nº of obs that we consider to apply uk and loocv
  
} else{
  #at the moment for pm10 and pm2.5
  name_pollutant  <- pollutant
  mod_constant    <- k[number]*1e09 #in the script of pm the units are kg/m3 but we are sure that they are worng, and they are in yg/m3
  min_obs         <- 5 #min nº of obs that we consider to apply uk and loocv
}

# Set temporal context ############################################################
# Description : 
#         Set temporal context will help us to  call the files based on the date, and subset the data. 
#         Also, it will determine the length of the loops based on time resolution of the data.

# Input : 
#        From configFile : date_time_begin, date_time_end.

# Output: 
#        date_times (data frame with dates and hours -no2&o3- ), dates (chr with yearmonthday), 
#        year_month (chr yearmonth), read_year (chr years), length_loop (int to determine number of temporal iterations for Ok, UK and LOOCV)

print("setting dates")
#date_times et fa un super vector d'una  columna, dia per dia, hora per hora des de l'inici fins a la data final p.ex: 2021-01-01 00:00:00
date_times <- data_frame_dates(date_time_begin = date_time_begin,
                               date_time_end = date_time_end)

#dates: Llista de caracters anymesdia p.ex: 20210101
dates <- unique(paste0(as.character(strftime(date_times[, 1], format="%Y", tz = "UTC")),
                       as.character(strftime(date_times[, 1], format="%m", tz = "UTC")),
                       as.character(strftime(date_times[, 1], format="%d", tz = "UTC"))))
#year_month: igual que date pero anymes p.ex: 202101
year_month <- unique(paste0(as.character(strftime(date_times[, 1], format="%Y", tz = "UTC")),
                            as.character(strftime(date_times[, 1], format="%m", tz = "UTC"))))

#For csv files, the name of obs files are referred to the year, so, at the moment we pick the first year in dates list:
#read_year <- substr(dates, 1, 4)[1]
read_year <- unique(substr(dates, 1, 4)) #with this command we will read the years in dates

#For daily data, we need a data frame with the days in format of %Y%m%d
daily_dates_format <- as.Date(dates, format = "%Y%m%d")
daily_dates        <- data.frame(date = daily_dates_format)

read_month <- as.integer(format(daily_dates$date, "%m"))

if(time_resolution=="hourly"){
  length_loop    <- nrow(date_times)
  pollutant_date <- date_times
} else if(time_resolution=="daily"){
  length_loop    <-length(dates)
  pollutant_date <- as.Date(dates, format = "%Y%m%d")
}

print(paste0("The length of the loop will be : ", length_loop))




# Load Local Stations ############################################################
# Description : 
#         In this script we will set up  a spatial points data frame called local stations, 
#         where we will have the code of the station and their coordinates based on the region bounded by the variable my_box.
#         my_box variable has two inputs from configFile (lon/lat _domain), are the maximum and minimum coordinates to determine which stations we pick
#         based on the region that we are studing

# Input : 
#        From confiFile : set_stations, file_format, obs_path (from ghost or gene), lon_domain, lat_domain

# Output: 
#     local_stations (Spatial points data frame with @data ($ref_obs, $type or station_name, $pollutant (NA)))   

print("Load Local Stations")
# local_stations: special points data frame with two cols. For example, for NO2 pollutant [NO2 and ref_obs-reference of station-], at the moment we have 72 stations in Cat for this pollutant 

if (set_stations){
  source(paste0(path_code,"set_stations.R"))
  print(paste0("Set stations load for ",pollutant, " using data from ", file_format, ": done"))
} else {
  local_stations <- readRDS(file = paste0(save_result,"local_stations_",file_format, "_", pollutant,".rds"))
  print(paste0("Read local stations for ",pollutant, " using data from ", file_format, ": done"))
}



# Load Observations ############################################################
# Description : 
#         Load obs prepares a data frame with the observations in each local station called df_obs_all, 
#         and another data frame with the substraction of the time related choosed configFile called df_obs_sub

# Input : 
#        From confiFile : set_observations, file_format, time_resolution, local_stations (from set_stations), name_pollutant,
#        length_loop, obs_path (path where the files of observations are), dates
#        obs_constant

# Output: 
#        df_obs_all & df_obs_sub

print("Load observations")

if (set_observations){
  source(paste0(path_code,"load_obs_cat.R"))
  print(paste0("Observations load for ",pollutant, " using data from ", file_format, ": done"))
  print("df_obs_all computed")
  
} else {
  df_obs_all <- readRDS(file = paste0(save_result,"df_obs_",file_format,"_",pollutant,".rds"))
  print("df_obs_all loaded from previous loading")
} 




# Subset observational data ############################################################

if(time_resolution=="hourly"){
  df_obs_sub <- time_subset(data = df_obs_all, time_var_name = "date",
                            date_time_begin = date_time_begin,
                            date_time_end = date_time_end)
  df_obs_sub$date <- as.POSIXct(df_obs_sub$date, tz="UTC")
  
} else if(time_resolution=="daily"){ 
  start_date  <- as.Date(as.character(dates[1]), format="%Y%m%d")
  end_date    <- as.Date(as.character(tail(dates,n=1)), format="%Y%m%d")
  df_obs_sub  <- subset(df_obs_all, date>=start_date & date <= end_date)

} else { 
  print("wrong time_resolution")
}  

# Check we don't have duplications in obs
if(any(duplicated(df_obs_sub$date))){
  print(paste0("Warning, there are duplicated observations on date :", df_obs_sub$date[duplicated(df_obs_sub$date)]))
  df_obs_sub <- distinct(df_obs_sub, date, .keep_all = TRUE)
  print("The df_obs_sub (alias=observations) has been modified to have the same length as the model")
}
# Check that we have all dates in observations
if(length(df_obs_sub$date) < length_loop & time_resolution=="hourly"){
  print("WARNING: There are missing hours/dates in observations")
  missing_dates     <- setdiff(date_times$date, df_obs_sub$date)
  missing_positions <- which(date_times$date %in% missing_dates)
  
  print(paste0("The day/hour missing in observations is: ",date_times$date[missing_positions]))
  
  empty_obs              <- matrix(NA, nrow=length_loop, ncol=0) 
  df_empty_obs           <- data.frame(empty_obs)
  df_empty_obs$date      <- date_times$date
  df_merged              <- full_join(df_empty_obs, df_obs_sub, by = "date", all.x = TRUE, all.y = TRUE)
  df_obs_sub             <- df_merged
  print("The observations [df_obs_sub] have been filled with NA on the missing dates")
  
}else if (length(df_obs_sub$date) != length_loop & time_resolution=="daily"){
  print("WARNING: There are missing hours/dates in observations. And the code is not developed for daily data yet!!")
}

print("Read observations & subsetting: done")


# Availability of AQMS ############################################################
# calcular availability para cada columna de df_obs_sub

if(set_stations_availability=="TRUE"){
  numero                  <- sapply(names(subset(df_obs_sub, select = -date)), function(x) 1 - sum(is.na(df_obs_sub[[x]]))/nrow(df_obs_sub))
  df_AQMS_availability    <- data.frame(t(numero), check.names=F)

  local_stations_availability              <- subset(local_stations, ref_obs %in% names(df_obs_sub))
  # convert df_AQMS into a named vector
  availability <- unlist(df_AQMS_availability)
  
  # extract availability values for ref_obs in local_stations_availability
  avail_values <- availability[local_stations_availability$ref_obs]
  
  # add availability values as a new column to local_stations_availability
  local_stations_availability$availability <- avail_values
  
  #local_stations_availability$availability <- df_AQMS_availability#[, local_stations_availability$ref_obs]

  saveRDS(object = local_stations_availability,file = paste0(save_result,"df_annual_AQMS_availability_",file_format,"_",pollutant,".rds"))
} else{
   local_stations_availability <- readRDS(paste0(save_result,"df_annual_AQMS_availability_",file_format,"_",pollutant,".rds"))
}


# Load Model ############################################################
# Description : 
#          The script calls the diferent files of the model, based on the dates choosed in configFile,
#          and prepares a matrix, called "model", with 3 dimentions, where we have lat, lon of each point of the model and the time related

# Input : 
#         From confiFile : set_model, dates, model_path, time_resolution, name_pollutant
#        mod_constant

# Output: 
#        model (matrix with 3 dimentions lat, lon and time), lat_model, lon_model [coords_model to save the coordinates of the model in case we read it]

print("Load Model")
if (set_model){
  # source(paste0(path_code,"load_caliope_cat.R"))
  source(paste0(path_code,"load_caliope.R"))
  print("Load Raw model load: done")
} else {
  model        <- readRDS(file = paste0(save_result,"model_",file_format, "_",pollutant,".rds"))
  coords_model <- readRDS(file = paste0(save_result,"coords_model_",file_format, "_",pollutant,".rds"))
  
  lon_model <- coords_model[, "lon_model"]
  lat_model <- coords_model[, "lat_model"]
  print("Preload raw model: done")
}




# Interpolate model to AQMS ############################################################
# Description : The script interpolates the points of the model, loaded in "load model" to the coordinates of the local stations (where we have observations)
# then the output is a data frame with the observations in the local stations, the model interpolated in them and the coordinates associated.

# Input : time_resolution, date_times, local_stations, model, length_loop, variogram_ok (if needed)

#Output: model_raw_at_AQMS_df, different_names

print("Interpolate model to AQMS")
if (execute_ok){
  source(paste0(path_code,"OK_interpolate_model_to_AQMS_v2.R"))
  print("Interpolate model to AQMS: computed")
} else {
  model_raw_at_AQMS_df <- readRDS(paste0(save_result,"model_raw_at_AQMS_",file_format, "_",pollutant,"_df.rds"))
  #maybe a substetting here: model_raw_at_AQMS_df=model_raw_at_AQMS_df[1:29,]
  #if(time_resolution=="daily"){
  #  different_names      <- readRDS(paste0(save_result,"different_local_station_",file_format, "_",pollutant,".rds"))
  #}
  print("Interpolate model to AQMS: loaded from previous files")
}


# GBM ############################################################
if(execute_gbm){
  source(paste0(path_code, "gbm_fill_obs_gaps.R"))
  print("GBM executed")
  if(time_resolution=="hourly"){
    df_gbm_sub <- time_subset(data = df_gbm_results, time_var_name = "date",
                              date_time_begin = date_time_begin,
                              date_time_end = date_time_end)
  } else if(time_resolution=="daily"){
    start_date  <- as.Date(as.character(dates[1]), format="%Y%m%d")
    end_date    <- as.Date(as.character(tail(dates,n=1)), format="%Y%m%d")
    df_gbm_sub  <- subset(df_gbm_results, date>=start_date & date <= end_date)
  
  } else {
    print("wrong time_resolution")
  }

} else if(execute_uk_gbm){
  df_gbm_sub <- readRDS(paste0(save_result, "gbm_gap_filling_",file_format,"_",pollutant,"_df.rds"))
  print("GBM read")
}




# Universal Kriging ####################################################
# Description : In this script we will evaluate the performance of universal kriging using leave one out cross validation
# Input : df_obs_sub, local_stations, length_loop, model_raw_at_AQMS_df, different_names
# Output: LOOCV_total_df, UK_total_df


print("Loop Universal Kriging")
if (execute_uk_loocv){
  if (execute_uk_gbm) {
    
    df_gbm_sub <- subset(df_gbm_sub, date>=start_date & date <= end_date)
    
    df_uk_observations <- df_gbm_sub
    print("Observations with gap filling will be used")
  } else {
    df_uk_observations <- df_obs_sub
    print("Observations without gap filling will be used")
  }
  
  source(paste0(path_code,"loop_uk_cat_v2.R"))
  print("Corrected model load: done")
  
} else {
  if(execute_uk){
#FIXME aixo ja no es guarda en un sol fitxer
    UK_total_df       <- readRDS(file = paste0(save_result, "UK_", file_format, "_", pollutant, "_df.rds"))
    #UK_total_GBM_df   <- readRDS(file = paste0(save_result, "UK_", file_format, "_", pollutant,"_gbm_", "df.rds"))
    
    print(paste("UK read it from obs", if (execute_uk_gbm) "with" else "without", "gap filling"))
  }
  if(execute_loocv){
    LOOCV_total_df     <- readRDS(file = paste0(save_result, "LOOCV_", file_format, "_", pollutant,"_", "df.rds"))
    #LOOCV_total_gbm_df <- readRDS(file = paste0(save_result, "LOOCV_", file_format, "_", pollutant,"_gbm_", "df.rds"))
    #print(paste("LOOCV read it from obs", if (execute_uk_gbm) "with" else "without", "gap filling")) #arreglar per 2 poss
  }
  else{
    print("WARNING: Not readed the corrected model or LOOCV")
  }
  print("Read corrected model: done")
}



# Save as netCDF ----
# system(paste("Rscript /esarchive/scratch/ccarnere/RScripts/scripts/aqs_tools/format_files/daily_rds_to_daily_nc.R", 
#              pollutant, read_year))



# Statistics ############################################################
# Description :

#    stadistics_uk_cat: Statistics_uk_cat compares the caliope raw model, uk correction results and burns model, where we extract the COE and MB of each. 
#    We can extract this stadistics for all stations concatenated, or the mean value for each hour for all stations. 

#    temporal_variation_cat: We plot the temporal variation for the mean of all stations, comparing raw model, uk corrected model and burnes. 
#    Finally, we plot and save the time variation for each station

#    time_series_uk_cat: idem as temporal variation but showing time series

# Input : 
#    all of them            :  date_times, df_obs_sub, local_stations, model_raw_at_AQMS_df, LOOCV_total_df, burnes_path, pollutant
#    stadistics_uk_cat      :  save_general_statistics
#    temporal_variation_cat :  save_temporal_variation
#    time_series_uk_cat     :  save_time_series

# Output: 
#    stadistics_uk_cat:  All_stations_hourly_no2_stats.png, All_stations_no2_stats.png
#    temporal_variation_cat: temporal_no2_mean_time_VAR.png, [station_code + "_no2_time_VAR.png"]
#    time_series_uk_cat:  [station_code + "_no2_month.png"]

options <- list(
  # "General Statistics"     = list("execute" = execute_general_st, "msg" = "General Stadistics","name"="stadistics"),
  "General Statistics"     = list("execute" = execute_general_st, "msg" = "General Stadistics","name"="statistics_v2.R"),
  "Temporal Variation"     = list("execute" = execute_temporal_var, "msg" = "Temporal variation","name"="temporal_variation_uk_cat.R"),
  "Time Series Evaluation" = list("execute" = execute_time_series, "msg" = "Time Series evaluation", "name"= "time_series_uk_cat.R")
)

if (execute_statistics) {
  
  print("Evaluation Statistics:")
  
  for (option in options) {
    if (option$execute) {
      source(path_code,option$name)
      print(paste0("- ", option$msg, ": done"))
    } else {
      print(paste0("- ", option$msg, ": not executed"))
    }
  }
  
} else {
  print("Evaluation Statistics not executed")
}


#Temps que triga en executar-se el fitxer
end_time <- Sys.time()
temps_total <- end_time-start_time
print(temps_total)
